from os import read
import pathlib, nidigital, logging
from time import sleep

from numpy import add
from .abstraction import ams_I2C_communication
from ams_ni_file_tools import compileDigipatSource 
from collections import namedtuple
from typing import List, Tuple

my_logger = logging.getLogger(__name__)

WRITE_CONTENT = """        
file_format_version 1.1;
timeset Idle_$name, Stop_$name, SourceStart_$name, Start_$name, BitTime_$name;

pattern I2C_write_$name (tmp_SDA_pin, tmp_SCL_pin)
{
source_d_replace(0, X)				Idle_$name			X		1;
source_start(downstream_data_$name)			Idle_$name			H		1;
									SourceStart_$name		H		1;
									Start_$name			0		1;
source								BitTime_$name			D		0;
source								BitTime_$name			D		0;
source								BitTime_$name			D		0;
source								BitTime_$name			D		0;
source								BitTime_$name			D		0;
source								BitTime_$name			D		0;
source								BitTime_$name			D		0;
source								BitTime_$name			D		0;
									BitTime_$name			L		0;
set_loop(reg0)						Idle_$name			X		1;
data:
source								BitTime_$name			D		0;
source								BitTime_$name			D		0;
source								BitTime_$name			D		0;
source								BitTime_$name			D		0;
source								BitTime_$name			D		0;
source								BitTime_$name			D		0;
source								BitTime_$name			D		0;
source								BitTime_$name			D		0;
end_loop(data)						BitTime_$name			L		0;
									Stop_$name			0		0;
halt								Idle_$name			X		1;
}
"""

READ_CONTENT = """        

file_format_version 1.1;
timeset Idle_$name, SourceStart_$name, Start_$name, BitTime_$name, Stop_$name;

pattern I2C_read_$name (tmp_SDA_pin, tmp_SCL_pin)
{
source_d_replace(0, X)				Idle_$name				X		1;
source_start(downstream_data_$name)			Idle_$name		H		1;
capture_start(upstream_data_$name)		Idle_$name		H		1;
									SourceStart_$name		H		1;
									Start_$name				0		1;
source								BitTime_$name				D		0;
source								BitTime_$name				D		0;
source								BitTime_$name				D		0;
source								BitTime_$name				D		0;
source								BitTime_$name				D		0;
source								BitTime_$name				D		0;
source								BitTime_$name				D		0;
source								BitTime_$name				D		0;
									BitTime_$name				L		0;
jump_if(seqflag0, readLastByte_$name)		Idle_$name		X		1;
set_loop(reg0)						Idle_$name				X		1;
readDataLoop_$name:
capture								BitTime_$name				V		0;
capture								BitTime_$name				V		0;
capture								BitTime_$name				V		0;
capture								BitTime_$name				V		0;
capture								BitTime_$name				V		0;
capture								BitTime_$name				V		0;
capture								BitTime_$name				V		0;
capture								BitTime_$name				V		0;
end_loop(readDataLoop_$name)	BitTime_$name		0		0;
readLastByte_$name:
capture								BitTime_$name				V		0;
capture								BitTime_$name				V		0;
capture								BitTime_$name				V		0;
capture								BitTime_$name				V		0;
capture								BitTime_$name				V		0;
capture								BitTime_$name				V		0;
capture								BitTime_$name				V		0;
capture								BitTime_$name				V		0;
									BitTime_$name				X		0;
capture_stop						Stop_$name				0		0;
halt								Idle_$name				X		1;
}
"""

READ_REG_CONTENT = """        

file_format_version 1.1;
timeset Idle_$name, SourceStart_$name, Start_$name, BitTime_$name, Stop_$name, rep_start_$name;

pattern I2C_rreg_$name (tmp_SDA_pin, tmp_SCL_pin)
{
source_d_replace(0, X)				Idle_$name			X		1;
capture_start(upstream_data_$name)		Idle_$name		H		1;
source_start(downstream_data_$name)			Idle_$name			H		1;
									SourceStart_$name		H		1;
									Start_$name			0		1;
source								BitTime_$name			D		0;
source								BitTime_$name			D		0;
source								BitTime_$name			D		0;
source								BitTime_$name			D		0;
source								BitTime_$name			D		0;
source								BitTime_$name			D		0;
source								BitTime_$name			D		0;
source								BitTime_$name			D		0;
									BitTime_$name			L		0;
									Idle_$name			X		1;
source								BitTime_$name			D		0;
source								BitTime_$name			D		0;
source								BitTime_$name			D		0;
source								BitTime_$name			D		0;
source								BitTime_$name			D		0;
source								BitTime_$name			D		0;
source								BitTime_$name			D		0;
source								BitTime_$name			D		0;
									BitTime_$name			L		0;
									rep_start_$name				1		0;
source								BitTime_$name				D		0;
source								BitTime_$name				D		0;
source								BitTime_$name				D		0;
source								BitTime_$name				D		0;
source								BitTime_$name				D		0;
source								BitTime_$name				D		0;
source								BitTime_$name				D		0;
source								BitTime_$name				D		0;
									BitTime_$name				L		0;
jump_if(seqflag0, readLastByte_$name)		Idle_$name		X		1;
set_loop(reg0)						Idle_$name				X		1;
readDataLoop_$name:
capture								BitTime_$name				V		0;
capture								BitTime_$name				V		0;
capture								BitTime_$name				V		0;
capture								BitTime_$name				V		0;
capture								BitTime_$name				V		0;
capture								BitTime_$name				V		0;
capture								BitTime_$name				V		0;
capture								BitTime_$name				V		0;
end_loop(readDataLoop_$name)	BitTime_$name		0		0;
readLastByte_$name:
capture								BitTime_$name				V		0;
capture								BitTime_$name				V		0;
capture								BitTime_$name				V		0;
capture								BitTime_$name				V		0;
capture								BitTime_$name				V		0;
capture								BitTime_$name				V		0;
capture								BitTime_$name				V		0;
capture								BitTime_$name				V		0;
									BitTime_$name				X		0;
capture_stop						Stop_$name				0		0;
halt								Idle_$name				X		1;
}
"""






class ams_pxi_I2C(ams_I2C_communication):
    """ams_pxi_I2C 
    This class implements the communication via I2C by using PXI digital pattern instrument
    for each communication channel a new object is created - each communication instance has a section in HW config file
    All the instances use the same pin map file - pin map file is project specific
    All I2C operations have dedicated pattern file - like write, read, write register, read register
    During the initialization the I2C operation's pattern files are generated for all instances - compile
    Timing information is held by instance property
    Pin signal level information is maintained in instance
    """
    timing_dict = {
        "BitTime": {
            "SDA": {
                "drive_on": 0.1,
                "data_on": 0.1,
                "return": 1,
                "drive_off": 1,
                "format": nidigital.DriveFormat.NR,
                "compare": 0.5
            },
            "SCL": {
                "drive_on": 0,
                "data_on": 0,
                "return": 0.5,
                "drive_off": 0.5,
                "format": nidigital.DriveFormat.RH,
                "compare": 0.25
            }
        },
        "Start": {
            "SDA": {
                "drive_on": 0.75,
                "data_on": 0.75,
                "return": 1,
                "drive_off": 1,
                "format": nidigital.DriveFormat.NR,
                "compare": 0.5
            },
            "SCL": {
                "drive_on": 0,
                "data_on": 0,
                "return": 0.5,
                "drive_off": 0.5,
                "format": nidigital.DriveFormat.NR,
                "compare": 0.25
            }
        },
        "rep_start": {
            "SDA": {
                "drive_on": 0,
                "data_on": 0,
                "return": 0.75,
                "drive_off": 1,
                "format": nidigital.DriveFormat.RL,
                "compare": 0.5
            },
            "SCL": {
                "drive_on": 0,
                "data_on": 0,
                "return": 0.5,
                "drive_off": 0.5,
                "format": nidigital.DriveFormat.NR,
                "compare": 0.25
            }
        },
        "Stop": {
            "SDA": {
                "drive_on": 0.1,
                "data_on": 0.1,
                "return": 1,
                "drive_off": 1,
                "format": nidigital.DriveFormat.NR,
                "compare": 0.5
            },
            "SCL": {
                "drive_on": 0,
                "data_on": 0,
                "return": 0.5,
                "drive_off": 0.5,
                "format": nidigital.DriveFormat.RH,
                "compare": 0.25
            }
        },
        "Idle": {
            "SDA": {
                "drive_on": 0,
                "data_on": 0,
                "return": 0.9,
                "drive_off": 0.9,
                "format": nidigital.DriveFormat.NR,
                "compare": 0.5
            },
            "SCL": {
                "drive_on": 0,
                "data_on": 0,
                "return": 0.9,
                "drive_off": 0.9,
                "format": nidigital.DriveFormat.NR,
                "compare": 0.5
            }
        },
        "SourceStart": {
            "SDA": {
                "drive_on": 0,
                "data_on": 0,
                "return": 0.9,
                "drive_off": 0.9,
                "format": nidigital.DriveFormat.NR,
                "compare": 0.5
            },
            "SCL": {
                "drive_on": 0,
                "data_on": 0,
                "return": 0.9,
                "drive_off": 0.9,
                "format": nidigital.DriveFormat.NR,
                "compare": 0.5
            }
        }
    }


    def __init__(self, device_dict: dict) -> object:
        """__init__ 
        This method initializes the communication interface based on the HW configuration file

        Args:
            device_dict (dict): dictionary containing the settings of the communication

        Returns:
            object: I2C communication object
        """
        self.comm_name = device_dict['Instrument_name']
        self.settings_dict = device_dict
        self.speed = device_dict['Details']['speed']
        self.bus_voltage = device_dict['Details']['bus_voltage']
        self.SCL_pin = device_dict['Details']['settings']['SCL']
        self.SDA_pin = device_dict['Details']['settings']['SDA']
        self.pin_map_file = device_dict['Details']['ni_pin_map_file']
        if 'instance' not in self.settings_dict['Details'].keys():
            my_logger.debug('Initialize new session')
            self.handler = nidigital.Session(
                resource_name=self.settings_dict['Details']['max_alias'],
                reset_device=False
            )
        else:
            my_logger.debug('Session is injected')
            self.handler = self.settings_dict['Details']['instance']

        self._my_timing_dict = self.timing_dict

        self._calcTimingParameters()
        self._setupTimesets()

        self._generateTimesetForSCL()
        self._generateTimesetForSDA()
        
        self._calculateLevelParameters()
        self._generateLevels()
        
        self._generatePatternFile()
        self._loadPatternFiles() 

        # self._createAddressBuffer()
        self._createDownstreamDataBuffer()
        self._createUpstreamDataBuffer()
        self.handler.commit()


    def reinit(self):
        self._my_timing_dict = self.timing_dict

        self._calcTimingParameters()
        self._setupTimesets()

        self._generateTimesetForSCL()
        self._generateTimesetForSDA()
        
        self._calculateLevelParameters()
        self._generateLevels()
        
        self._generatePatternFile()
        try:
            self._loadPatternFiles() 
        except nidigital.Error:
            pass

        # self._createAddressBuffer()
        self._createDownstreamDataBuffer()
        self._createUpstreamDataBuffer()
        self.handler.commit()
        



    def _calcTimingParameters(self):
        """_calcPeriod 
        Calculates the period from speed
        Calculates all the timing parameters based on 
        """
        self.period = 1 / self.speed
        my_logger.debug(f'Period: {self.period * 1000000} us')

        self.timing_parameters = {}

        for timing in self._my_timing_dict.keys():
            # my_logger.debug(timing)
            timing_dict = {}
            for pin in self._my_timing_dict[timing].keys():
                # my_logger.debug(pin)
                pin_dict = {}
                for factor in self._my_timing_dict[timing][pin].keys():
                    # my_logger.debug(factor)
                    if factor != 'format':
                        pin_dict[factor] = self._my_timing_dict[timing][pin][factor] * self.period
                        # my_logger.debug(f'{timing}-{pin}-{factor}: {pin_dict[factor] * 1000000} us')
                    else: \
                        pin_dict[factor] = self._my_timing_dict[timing][pin][factor]
                    # my_logger.debug(pin_dict[factor])
                timing_dict[pin] = pin_dict
            # my_logger.debug(timing_dict[pin])
            self.timing_parameters[f'{timing}_{self.comm_name}'] = timing_dict


    def _setupTimesets(self):
        for timing in self.timing_parameters.keys():
            # my_logger.debug(type(self.handler))
            self.handler.create_time_set(timing)


    def _generateTimesetForSDA(self):
        """_generateTimesetForSDA 
        This protected method calculates the timing parameters for the data
        """
        for timing in self.timing_parameters.keys():
            # my_logger.debug(f'Grab data for {timing}')
            drive_on = self.timing_parameters[timing]['SDA']['drive_on']
            data_on = self.timing_parameters[timing]['SDA']['data_on']
            return_t = self.timing_parameters[timing]['SDA']['return']
            p_format = self.timing_parameters[timing]['SDA']['format']
            drive_off = self.timing_parameters[timing]['SDA']['drive_off']
            compare_t = self.timing_parameters[timing]['SDA']['compare']
            # my_logger.debug(f'configuring {timing} for SDA')
            self.handler.pins[self.SDA_pin].configure_time_set_drive_edges(
                time_set_name=timing,
                format=p_format,
                drive_on_edge=drive_on,
                drive_data_edge=data_on,
                drive_return_edge=return_t,
                drive_off_edge=drive_off
            )
            self.handler.pins[self.SDA_pin].configure_time_set_compare_edges_strobe(
                time_set_name=timing,
                strobe_edge=compare_t
            )
            self.handler.configure_time_set_period(
                time_set_name=timing,
                period=self.period
            )


    def _generateTimesetForSCL(self):
        """_generateTimesetForSCL 
        This method calculates all the timint parameters for the clock
        """
        for timing in self.timing_parameters.keys():
            # my_logger.debug(f'Grab data for {timing}')
            drive_on = self.timing_parameters[timing]['SCL']['drive_on']
            data_on = self.timing_parameters[timing]['SCL']['data_on']
            return_t = self.timing_parameters[timing]['SCL']['return']
            p_format = self.timing_parameters[timing]['SCL']['format']
            drive_off = self.timing_parameters[timing]['SCL']['drive_off']
            compare_t = self.timing_parameters[timing]['SCL']['compare']
            # my_logger.debug(f'configuring {timing} for SCL')
            self.handler.pins[self.SCL_pin].configure_time_set_drive_edges(
                time_set_name=timing,
                format=p_format,
                drive_on_edge=drive_on,
                drive_data_edge=data_on,
                drive_return_edge=return_t,
                drive_off_edge=drive_off
            )
            self.handler.pins[self.SCL_pin].configure_time_set_compare_edges_strobe(
                time_set_name=timing,
                strobe_edge=compare_t
            )
            self.handler.configure_time_set_period(
                time_set_name=timing,
                period=self.period
            )

    
    def _generateLevels(self):
        """_generateLevels 
        This method calculates the voltage level information
        """
        self.handler.pins[self.SDA_pin].configure_voltage_levels(
            vih=self.voltage_parameters['level_high'],
            vil=self.voltage_parameters['level_low'],
            voh=self.voltage_parameters['comp_high'],
            vol=self.voltage_parameters['comp_low'],
            vterm=0
        )
        self.handler.pins[self.SCL_pin].configure_voltage_levels(
            vih=self.voltage_parameters['level_high'],
            vil=self.voltage_parameters['level_low'],
            voh=self.voltage_parameters['comp_high'],
            vol=self.voltage_parameters['comp_low'],
            vterm=0
        )


    def _calculateLevelParameters(self):
        """_calculateParameters 
        calculates the voltage levels based on the bus voltage
        """
        self.voltage_parameters = {
            "level_high": self.bus_voltage,
            "level_low": 0,
            "comp_high": self.bus_voltage * 0.8,
            "comp_low": self.bus_voltage * 0.2
        }


    def _generatePatternFile(self):
        """_generatePatternFile 
        This method generates the pattern file for all the different I2C operations and saves them into the same directory where the pin map file can be found
        """
        self._generateWrite()
        writeOK, self.writePatternFile = compileDigipatSource(
            sourcePath=self.writeSrcPath,
            pinmapPath=self.pin_map_file
        )
        self._generateRead()
        readOK, self.readPatternFile = compileDigipatSource(
            sourcePath=self.readSrcPath,
            pinmapPath=self.pin_map_file
        )
        self._generateReadReg()
        readOK, self.readRegPatternFile = compileDigipatSource(
            sourcePath=self.readRegSrcPath,
            pinmapPath=self.pin_map_file
        )
        return writeOK and readOK


    def __getWorkingDirectory(self) -> str:
        """__getWorkingDirectory 
        Returns the directory where the pin map file is located

        Returns:
            str: pin map file directory
        """
        return str(pathlib.Path(self.pin_map_file).parent.absolute())


    def _generateWrite(self):
        """_generateWrite 
        generates the src file for the write operation
        """
        self.writeSrcPath = f'{self.__getWorkingDirectory()}\\{self.comm_name}_I2C_write.digipatsrc'
        with open(self.writeSrcPath, 'w') as write_pattern_file:
            pattern_string = self._setPinsInTemaplateString(WRITE_CONTENT)
            pattern_string = self._setPatternName(pattern_string)
            write_pattern_file.write(pattern_string)


    def _generateRead(self):
        """_generateRead 
        generates the src file for the read operation
        """
        self.readSrcPath = f'{self.__getWorkingDirectory()}\\{self.comm_name}_I2C_read.digipatsrc'
        with open(self.readSrcPath, 'w') as pattern_file:
            pattern_string = self._setPinsInTemaplateString(READ_CONTENT)
            pattern_string = self._setPatternName(pattern_string)
            pattern_file.write(pattern_string)


    def _generateReadReg(self):
        """_generateReadReg 
        generates the src file for the read register operation
        """
        self.readRegSrcPath = f'{self.__getWorkingDirectory()}\\{self.comm_name}_I2C_rreg.digipatsrc'
        with open(self.readRegSrcPath, 'w') as pattern_file:
            pattern_string = self._setPinsInTemaplateString(READ_REG_CONTENT)
            pattern_string = self._setPatternName(pattern_string)
            pattern_file.write(pattern_string)


    def _setPinsInTemaplateString(self, content: str) -> str:
        """_setPinsInTemaplateString 
        This method replaces the pin names in a given template content

        Args:
            content (str): the content of the template

        Returns:
            str: modified template string
        """
        new_content = content.replace('tmp_SDA_pin', self.SDA_pin)
        new_content = new_content.replace('tmp_SCL_pin', self.SCL_pin)
        return new_content


    def _setPatternName(self, content: str) -> str:
        new_content = content.replace('$name', self.comm_name)
        return new_content


    def _createAddressBuffer(self):
        self.handler.channels[self.SDA_pin].create_source_waveform_serial(
            waveform_name=f'address_{self.comm_name}',
            bit_order=nidigital.BitOrder.MSB,
            sample_width=8,
            data_mapping=nidigital.SourceDataMapping.BROADCAST
        )


    def _createDownstreamDataBuffer(self):
        self.handler.channels[self.SDA_pin].create_source_waveform_serial(
            waveform_name=f'downstream_data_{self.comm_name}',
            bit_order=nidigital.BitOrder.MSB,
            sample_width=8,
            data_mapping=nidigital.SourceDataMapping.BROADCAST
        )


    def _createUpstreamDataBuffer(self):
        self.handler.channels[self.SDA_pin].create_capture_waveform_serial(
            waveform_name=f'upstream_data_{self.comm_name}',
            bit_order=nidigital.BitOrder.MSB,
            sample_width=8
        )


    def _loadPatternFiles(self):
        my_logger.debug(self.writePatternFile)
        my_logger.debug(self.readPatternFile)
        my_logger.debug(self.readRegPatternFile)
        self.handler.load_pattern(self.writePatternFile)
        self.handler.load_pattern(self.readPatternFile)
        self.handler.load_pattern(self.readRegPatternFile)


    def _setAddress(self, address: int):
        self.handler.write_source_waveform_broadcast(f'address_{self.comm_name}', [address])


    def _setDownstreamData(self, data: list):
        self.handler.write_source_waveform_broadcast(f'downstream_data_{self.comm_name}', data)


    def _setMessageLength(self, message_length: int):
        self.handler.write_sequencer_register(nidigital.SequencerRegister.REGISTER0, message_length)


    def I2C_init(self, device_name):
        """
        This method initializes the I2C communication
        """
        pass


    def I2C_configure(
        self,
        speed: int=None,
        bus_voltage: float=None,
        settings_dict: dict={}
    ):
        """I2C_configure 
        This method configures the communication

        Args:
            speed (int): communication speed
            bus_voltage (float): bus voltage in V
            settings_dict (dict, optional): defines any implementation setting that may be necessary. Defaults to {}.
        """
        self.speed = speed if speed != None else self.speed
        self.bus_voltage = bus_voltage if bus_voltage != None else self.bus_voltage

        my_logger.debug(settings_dict.keys())
        if 'timing_dict' in settings_dict.keys():
            my_logger.debug('Timing dict configured')
            self._my_timing_dict = settings_dict['timing_dict']

        self._calcTimingParameters()
        self._generateTimesetForSCL()
        self._generateTimesetForSDA()
        self._calculateLevelParameters()
        self._generateLevels()
        self.handler.commit()


    @classmethod
    def _generatePayload(self, payload: list) -> list:
        final_payload = []
        for index in range(14):
            try:
                final_payload.append(payload[index])
            except IndexError:
                final_payload.append(0x00)
        return final_payload


    @classmethod
    def _generateAddressByte(self, device_address: int, read: bool=False) -> int:
        """_generateAddressByte 
        Generates the address byte from device address.
        It uses the information whether read or write operation is going to be done.
        It converts the 7 bit device address to 8 bit address byte.
        In case of:
        - read operation - LSB is set to 1
        - write operation - LSB is set to 0
        all the rest of the bits are shifted left

        Args:
            device_address (int): device address
            read (bool, optional): decides whether the operation is read or write. Defaults to False.

        Returns:
            int: address byte
        """
        # my_logger.debug(f'Device address: {device_address:02x}')
        # my_logger.debug(f'Device address: {device_address:08b}')
        address_byte = ( device_address << 1 ) + (1 * int(read))
        # my_logger.debug(f'Address byte: {address_byte:02x}')
        # my_logger.debug(f'Address byte: {address_byte:08b}')
        return address_byte


    def I2C_write(
        self, 
        address: int, 
        payload: List[int]
        ) -> bool:
        """I2C_write 
        This method send I2C message in writing mode

        Args:
            address (int): 8 bit address of the IC (7 bit address << 1)
            payload (List[int]): A list of the payload bytes to be transmitted

        Returns:
            bool: True on success, False if the Slave doesn't respond
        """
        my_logger.debug(f'Writing to {address:02x}: {payload}')
        my_payload = [self._generateAddressByte(address), *payload]
        my_logger.debug(f'raw msg content: {my_payload}')
        self._setDownstreamData(self._generatePayload(my_payload))
        self._setMessageLength(len(my_payload) - 1)
        self.handler.commit()
        res = self.handler.burst_pattern(f'I2C_write_{self.comm_name}')

        return res


    def I2C_read(
        self, 
        address: int, 
        read_count: int
        ) -> Tuple[bool, List[int]]:
        """I2C_read 
        This method send I2C message in reading mode

        Args:
            address (int): 8 bit address of the slave (7 bit address << 1)
            read_count (int): number of bytes to read

        Returns:
            Tuple[bool, List[int]]: First element: True on success, Second element: The read bytes as list
        """
        if (read_count == 1):
            self.handler.write_sequencer_flag(nidigital.SequencerFlag.FLAG0, True)
        else:
            self.handler.write_sequencer_flag(nidigital.SequencerFlag.FLAG0, False)
        self._setDownstreamData(self._generatePayload([self._generateAddressByte(address, True)]))
        self._setMessageLength(read_count)
        self.handler.commit()
        stat = self.handler.burst_pattern(f'I2C_read_{self.comm_name}')
        my_logger.debug(stat)
        return stat, self._decodeCapture(read_count)

    
    def _decodeCapture(self, read_count: int) -> list:
        result = self.handler.fetch_capture_waveform(
            f'upstream_data_{self.comm_name}', read_count
        )
        result = result[0].tobytes()[0::4]
        result = [int(byte) for byte in result]
        return result


    def I2C_register_write(
        self, 
        device_address: int, 
        register_address: int, 
        payload: List[int]
        ) -> bool:
        """I2C_register_write 
        Write bytes to the memory of a I2C Slave with EEPROM-like interface

        Args:
            device_address (int): 8 bit address of the device (7 bit address << 1)
            register_address (int): 8 bit address of the first byte to write to
            payload (List[int]): List of bytes to write to the device

        Returns:
            bool: True on success
        """
        return self.I2C_write(
            address=device_address,
            payload=[register_address, *payload]
        )


    def I2C_register_read(
        self, 
        device_address: int, 
        register_address: int, 
        read_count: int
        ) -> Tuple[bool, List[int]]:
        """I2C_register_read 
        Read bytes from the memory of a I2C Slave with EEPROM-like interface

        Args:
            device_address (int): 8 bit address of the device (7 bit address << 1)
            register_address (int): 8 bit address of the first byte to read
            read_count (int): Number of bytes to read

        Tuple[bool, List[int]]: First element: True on success, Second element: The read bytes as list
        """
        if (read_count == 1):
            self.handler.write_sequencer_flag(nidigital.SequencerFlag.FLAG0, True)
        else:
            self.handler.write_sequencer_flag(nidigital.SequencerFlag.FLAG0, False)

        my_payload = self._generatePayload([
            self._generateAddressByte(device_address),
            register_address,
            self._generateAddressByte(device_address, True)
        ])

        self._setDownstreamData(my_payload)
        self._setMessageLength(read_count)
        self.handler.commit()
        stat = self.handler.burst_pattern(f'I2C_rreg_{self.comm_name}')
        my_logger.debug(stat)
        return stat, self._decodeCapture(read_count)
